from amuse.test import amusetest

from amuse.units import units
from amuse.units import constants
from amuse.units import nbody_system
import numpy
import time
import sys

from amuse.datamodel import Particles
from amuse.datamodel import Particle
from amuse.couple import encounters


#codes to use
from amuse.community.kepler.interface import Kepler
from amuse.community.smalln.interface import SmallN

def new_binary(
        mass1, mass2, semi_major_axis,
        eccentricity = 0, keyoffset = 1,
        is_at_periapsis = True,
        G =  nbody_system.G 
    ):
    total_mass = mass1 + mass2
    mass_fraction_particle_1 = mass1 / (total_mass)

    binary = Particles(keys=list(range(keyoffset, keyoffset+2)))
    binary[0].mass = mass1
    binary[1].mass = mass2
    binary.child1 = None
    binary.child2 = None
    
    mu = G * total_mass

    if is_at_periapsis:
        velocity = numpy.sqrt( mu / semi_major_axis  * ((1.0 + eccentricity)/(1.0 - eccentricity)))
        radius   = semi_major_axis * (1.0 - eccentricity)
    else:
        velocity = numpy.sqrt( mu / semi_major_axis  * ((1.0 - eccentricity)/(1.0 + eccentricity)))
        radius   = semi_major_axis * (1.0 + eccentricity)
        
    binary[0].position = ((1.0 - mass_fraction_particle_1) * radius * [1.0,0.0,0.0])
    binary[1].position = -(mass_fraction_particle_1 * radius * [1.0,0.0,0.0])

    binary[0].velocity = ((1.0 - mass_fraction_particle_1) * velocity * [0.0,1.0,0.0])
    binary[1].velocity = -(mass_fraction_particle_1 * velocity * [0.0,1.0,0.0])   

    return binary
       
class TryHandleEncounter(encounters.AbstractHandleEncounter, encounters.SelectNeighboursByDistanceMixin):
    
    def __init__(self,
        kepler_code,
        G = nbody_system.G
    ):
        encounters.AbstractHandleEncounter.__init__(
            self,
            kepler_code,
            G
        )
        encounters.SelectNeighboursByDistanceMixin.__init__(self)
    
class TestAbstractHandleEncounter(amusetest.TestWithMPI):
    
    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x
        
    def test1(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(5)
        particles_in_field.mass = [1,2,3,4,5] | nbody_system.mass
        particles_in_field[0].position = [2,0,0] | nbody_system.length
        particles_in_field[1].position = [1.5,0,0] | nbody_system.length
        particles_in_field[2].position = [0.5,1,0] | nbody_system.length
        particles_in_field[3].position = [0.5,-0.5,0] | nbody_system.length
        particles_in_field[4].position = [0,0,2] | nbody_system.length
        particles_in_field.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.large_scale_of_particles_in_the_encounter, 1.0 | nbody_system.length)
        self.assertAlmostRelativeEqual(x.small_scale_of_particles_in_the_encounter, 3.0 | nbody_system.length)
        
        self.assertEqual(len(x.particles_close_to_encounter), 3)
        self.assertAlmostRelativeEqual(x.particles_close_to_encounter.mass, [2,3,4] | nbody_system.mass)
        
    def test2(self):
        
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(keys=(5,))
        particles_in_field.mass = 2 | nbody_system.mass
        particles_in_field.position = [-0.5,0,0] | nbody_system.length
        particles_in_field.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        particles_in_multiples = Particles()
        particles_in_multiples.add_particle(particles_in_encounter[0])
        multiple = particles_in_multiples[0]
        multiple.components = Particles(keys = (3,4))
        multiple.components.mass = 0.5 | nbody_system.mass
        multiple.components[0].position = [0,0.1,0] | nbody_system.length
        multiple.components[1].position = [0,-0.1,0] | nbody_system.length
        multiple.components[0].velocity = [0,0,0.2] | nbody_system.speed
        multiple.components[1].velocity = [0,0.1,-0.2] | nbody_system.speed
        multiple.components.child1 = None
        multiple.components.child2 = None
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        x.existing_multiples.add_particles(particles_in_multiples)
        
        x.execute()
        
        self.assertEqual(len(x.all_singles_in_encounter), 3)
        self.assertEqual(len(x.all_singles_in_evolve), 4)
        
        
        self.assertAlmostRelativeEqual(x.all_singles_in_evolve.mass, [0.5, 0.5, 1.0, 2.0]| nbody_system.mass)
        
        self.assertEqual(len(x.released_singles), 2)
        print(x.particles_after_encounter)
        child1 = x.particles_after_encounter[-2]
        child2 = x.particles_after_encounter[-1]
        print(child1.position)
        print(child2.position)
        self.assertAlmostRelativeEqual(child1.position, [1,0.1,0]| nbody_system.length)
        self.assertAlmostRelativeEqual(child1.velocity, [0,0.5,0.2]| nbody_system.speed)
        self.assertAlmostRelativeEqual(child2.position, [1,-0.1,0]| nbody_system.length)
        self.assertAlmostRelativeEqual(child2.velocity, [0,0.6,-0.2]| nbody_system.speed)
        
        
    def test3(self):
        
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(keys=(2,4,5,6,7,))
        particles_in_field.mass = [1,2,3,4,5] | nbody_system.mass
        particles_in_field[0].position = [2,0,0] | nbody_system.length
        particles_in_field[1].position = [1.5,0,0] | nbody_system.length
        particles_in_field[2].position = [0.5,1,0] | nbody_system.length
        particles_in_field[3].position = [0.5,-0.5,0] | nbody_system.length
        particles_in_field[4].position = [0,0,2] | nbody_system.length
        particles_in_field.velocity = [0,0,0]  | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
    
    
        simple_binary = new_binary(
            1 | nbody_system.mass, 
            1 | nbody_system.mass, 
            0.2 | nbody_system.length
        )
                
        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None
            
            root_particle = particles.add_particle(Particle(
                key = 10,
                mass = 2.0 | nbody_system.mass, 
                position = particles[0:2].center_of_mass(),
                velocity = particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position
            
            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity
        
        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda : 1
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        
        self.assertAlmostRelativeEqual(multiple.components[0].position, simple_binary[0].position)
        self.assertAlmostRelativeEqual(multiple.components[1].position, simple_binary[1].position) 
    
    
        
    def test4(self):
        
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(keys=(3,4,5,6,7,))
        particles_in_field.mass = [1,2,3,4,5] | nbody_system.mass
        particles_in_field[0].position = [2,0,0] | nbody_system.length
        particles_in_field[1].position = [1.5,0,0] | nbody_system.length
        particles_in_field[2].position = [0.5,1,0] | nbody_system.length
        particles_in_field[3].position = [0.5,-0.5,0] | nbody_system.length
        particles_in_field[4].position = [0,0,2] | nbody_system.length
        particles_in_field.velocity = [0,0,0]  | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        binaries = Particles(keys=(20,))
        binaries[0].child1 = particles_in_encounter[0]
        binaries[0].child2 = particles_in_encounter[1]
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        x.existing_binaries.add_particles(binaries)
        
        
        simple_binary = new_binary(
            1 | nbody_system.mass, 
            1 | nbody_system.mass, 
            0.2 | nbody_system.length
        )
                
        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None
            
            root_particle = particles.add_particle(Particle(
                key = 10,
                mass = 2.0 | nbody_system.mass, 
                position = particles[0:2].center_of_mass(),
                velocity = particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position
            
            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity
        
        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda : 1
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.updated_binaries), 1)
        self.assertEqual(x.updated_binaries[0], binaries[0])



    def test5(self):
        
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(keys=(2,4,5,6,7,))
        particles_in_field.mass = [1,2,3,4,5] | nbody_system.mass
        particles_in_field[0].position = [2,0,0] | nbody_system.length
        particles_in_field[1].position = [1.5,0,0] | nbody_system.length
        particles_in_field[2].position = [0.5,1,0] | nbody_system.length
        particles_in_field[3].position = [0.5,-0.5,0] | nbody_system.length
        particles_in_field[4].position = [0,0,2] | nbody_system.length
        particles_in_field.velocity = [0,0,0]  | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.parameters.hard_binary_factor = 1
        x.small_scale_factor = 1

        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        
        simple_binary = new_binary(
            1 | nbody_system.mass, 
            1 | nbody_system.mass, 
            2 | nbody_system.length
        )
                
        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None
            
            root_particle = particles.add_particle(Particle(
                key = 10,
                mass = 2.0 | nbody_system.mass, 
                position = particles[0:2].center_of_mass(),
                velocity = particles[0:2].center_of_mass_velocity(),
            ))
            root_particle.child1 = particles[0]
            root_particle.child2 = particles[1]
            particles[0].position = simple_binary[0].position + root_particle.position
            particles[1].position = simple_binary[1].position + root_particle.position
            
            particles[0].velocity = simple_binary[0].velocity + root_particle.velocity
            particles[1].velocity = simple_binary[1].velocity + root_particle.velocity
        
        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda : 1
        
        x.execute()
        
        # no multiples as the binary is larger than the 
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        

    def test6(self):
        
        particles_in_encounter = Particles(keys=(1,2,3))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter[2].position = [0,0.5,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.small_scale_factor = 1.0
        
        simple_binary_1 = new_binary(
            1 | nbody_system.mass, 
            1 | nbody_system.mass, 
            0.4 | nbody_system.length
        )
        simple_binary_top = new_binary(
            2 | nbody_system.mass, 
            1 | nbody_system.mass, 
            2 | nbody_system.length
        )
                
        def evolve_singles_in_encounter_until_end_state():
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            particles.child1 = None
            particles.child2 = None
            
            inner_binary_particle = particles.add_particle(Particle(
                key = 10,
                mass = 2.0 | nbody_system.mass, 
                position = particles[0:2].center_of_mass(),
                velocity = particles[0:2].center_of_mass_velocity(),
            ))
            inner_binary_particle.child1 = particles[0]
            inner_binary_particle.child2 = particles[1]
            particles[0].position = simple_binary_1[0].position + inner_binary_particle.position
            particles[1].position = simple_binary_1[1].position + inner_binary_particle.position
            
            particles[0].velocity = simple_binary_1[0].velocity + inner_binary_particle.velocity
            particles[1].velocity = simple_binary_1[1].velocity + inner_binary_particle.velocity
            
            root_particle = particles.add_particle(Particle(
                key = 11,
                mass = 3.0 | nbody_system.mass, 
                position = particles.center_of_mass(),
                velocity = particles.center_of_mass_velocity(),
            ))
            
            root_particle.child1 = inner_binary_particle
            root_particle.child2 = particles[2]
            inner_binary_particle.position = simple_binary_top[0].position + root_particle.position
            particles[2].position = simple_binary_top[1].position + root_particle.position
            
            inner_binary_particle.velocity = simple_binary_top[0].velocity + root_particle.velocity
            particles[2].velocity = simple_binary_top[1].velocity + root_particle.velocity
            
            
            
        
        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda : 1
        
        x.execute()
        
        # no multiples as the binary is larger than the 
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
    
    
    
    def test7(self):
        
        particles_in_encounter = Particles(keys=(1,2,3))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter[2].position = [0,0.5,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        multiples = Particles()
        particle_in_multiples = multiples.add_particle( particles_in_encounter[2])
        particle_in_multiples.components = Particles(keys=(4,5))
        particle_in_multiples.components.mass = 1 | nbody_system.mass
        particle_in_multiples.components[0].position = [0,0,0.2] | nbody_system.length
        particle_in_multiples.components[1].position = [0,0,-0.2] | nbody_system.length
        particle_in_multiples.components.velocity = [0,0.0,0] | nbody_system.speed
        particle_in_multiples.components.child1 = None
        particle_in_multiples.components.child2 = None
        x.existing_multiples.add_particles(multiples) 
        
                
        def evolve_singles_in_encounter_until_end_state():
            
            particles = x.singles_and_multiples_after_evolve
            particles.add_particles(x.all_singles_in_encounter)
            print(particles)
            particles.child1 = None
            particles.child2 = None
        
            
            
        
        x.evolve_singles_in_encounter_until_end_state = evolve_singles_in_encounter_until_end_state
        x.determine_structure_of_the_evolved_state = lambda : 1
        
        x.execute()
        
        # no multiples as the binary is larger than the 
        # hard binary scale
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.dissolved_multiples), 1)
        self.assertEqual(len(x.released_singles), 2)
        self.assertTrue(particle_in_multiples.components[0] in x.released_singles)
        self.assertTrue(particle_in_multiples.components[1] in x.released_singles)
        
    
    def xtest8(self):
        particles_in_encounter = Particles(3)
        particles_in_encounter.mass = 1. | nbody_system.mass
        particles_in_encounter[0].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[2].position = [2.,0.,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.5,0.] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 1., 0.2] | nbody_system.length
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.large_scale_of_particles_in_the_encounter, 2.0 | nbody_system.length)
        self.assertAlmostRelativeEqual(x.small_scale_of_particles_in_the_encounter, 1.5 | nbody_system.length)
    
    
    
    def test8b(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0,0,0] | nbody_system.length
        particles_in_encounter[1].position = [1,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
       
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.initial_potential_in_field,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        
    
    def test9(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0,0,0] | nbody_system.length
        particles_in_encounter[1].position = [1,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
       
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.initial_potential_in_field,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)
        
        
    def test10(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0,0,0] | nbody_system.length
        particles_in_encounter[1].position = [1,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass = 1 | nbody_system.mass, 
            position = [5,0,0]  | nbody_system.length,
            velocity = [0,0,0]  | nbody_system.speed,
            radius = 0 | nbody_system.length 
        ))
       
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   -0.45 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)
        
    def test11(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0,0,0] | nbody_system.length
        particles_in_encounter[1].position = [1,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length
        
        particles_in_multiples = Particles()
        particles_in_multiples.add_particle(particles_in_encounter[0])
        multiple = particles_in_multiples[0]
        multiple.components = Particles(keys = (3,4))
        multiple.components.mass = 0.5 | nbody_system.mass
        multiple.components[0].position = [0,0.1,0] | nbody_system.length
        multiple.components[1].position = [0,-0.1,0] | nbody_system.length
        multiple.components[0].velocity = [0,0,0.2] | nbody_system.speed
        multiple.components[1].velocity = [0,0.1,-0.2] | nbody_system.speed
        multiple.components.child1 = None
        multiple.components.child2 = None
        
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass = 1 | nbody_system.mass, 
            position = [5,0,0]  | nbody_system.length,
            velocity = [0,0,0]  | nbody_system.speed,
            radius = 0 | nbody_system.length 
        ))
        x.existing_multiples.add_particles(particles_in_multiples)
       
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   -0.45 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -0.75 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -1.912495| nbody_system.energy,4)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0.06500400 | nbody_system.energy,4)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  -1.2275 | nbody_system.energy,4)
        self.assertAlmostRelativeEqual(x.final_energy, -1.9124959996 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.delta_phi_2, 0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  0 | nbody_system.energy)
        
    def test12(self):
        particles_in_encounter = Particles(2)
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [0,0,0] | nbody_system.length
        particles_in_encounter[1].position = [1,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.5,0] | nbody_system.speed
        particles_in_encounter.radius = [0.5, 0.5] | nbody_system.length
        
        
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particle(Particle(
            mass = 0.1 | nbody_system.mass, 
            position = [1.5,0,0]  | nbody_system.length,
            velocity = [0,0,0]  | nbody_system.speed,
            radius = 1 | nbody_system.length 
        ))
       
        
        x.execute()
        
        self.assertAlmostRelativeEqual(x.initial_potential_in_field,   0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_energy, -1.016666 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.initial_singles_energy, -1.016666 | nbody_system.energy, 4)
        self.assertAlmostRelativeEqual(x.delta_phi_1,  0 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy,  0 | nbody_system.energy)
        
    
    def test13(self):
        x = TryHandleEncounter(
            G = nbody_system.G,
            kepler_code = self.new_kepler()
        )
        
        x.parameters.hard_binary_factor = 1
        self.assertEqual(x.parameters.hard_binary_factor, 1)
        self.assertEqual(x.parameters.scatter_factor, 10)
        x.parameters.scatter_factor = 2
        self.assertEqual(x.parameters.scatter_factor, 2)
        self.assertEqual(x.get_scatter_factor(), 2)
        
class TestKeplerOrbits(amusetest.TestWithMPI):
    
    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x
        
    def test1(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary( 
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
        )
        semimajor_axis, eccentricity = x.get_semimajor_axis_and_eccentricity_for_binary_components(
            binary[0],
            binary[1]
        )
        self.assertAlmostRelativeEquals(semimajor_axis,  1.2 | nbody_system.length)
        self.assertAlmostRelativeEquals(eccentricity,  0)
        
    def test2(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary( 
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            2.0 | nbody_system.length,
            0.5,
            is_at_periapsis = False
            
        )
        binary.position += [0.1,0.2,0.3]  | nbody_system.length
        binary.velocity += [0.4,0.5,0.6]  | nbody_system.speed
        
        dpos, dvel = x.compress_binary(
            binary,
            1.2 | nbody_system.length
        )
        
        center_of_mass_before = binary.center_of_mass()
        center_of_mass_velocity_before = binary.center_of_mass_velocity()
        
        binary.position += dpos
        binary.velocity += dvel
        
        center_of_mass_after = binary.center_of_mass()
        center_of_mass_velocity_after = binary.center_of_mass_velocity()
        
        self.assertAlmostRelativeEquals(center_of_mass_before, center_of_mass_after)
        self.assertAlmostRelativeEquals(center_of_mass_velocity_before, center_of_mass_velocity_after)
        separation = (binary[0].position  - binary[1].position).length()
        self.assertAlmostRelativeEquals( separation, 1.2 | nbody_system.length)
    
    def test3(self):
        x = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary( 
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            2.0 | nbody_system.length,
            0.5,
            is_at_periapsis = True
            
        )
        binary.position += [0.1,0.2,0.3]  | nbody_system.length
        binary.velocity += [0.4,0.5,0.6]  | nbody_system.speed
        
        dpos, dvel = x.expand_binary(
            binary,
            2.5 | nbody_system.length
        )
        
        center_of_mass_before = binary.center_of_mass()
        center_of_mass_velocity_before = binary.center_of_mass_velocity()
        
        binary.position += dpos
        binary.velocity += dvel
        
        center_of_mass_after = binary.center_of_mass()
        center_of_mass_velocity_after = binary.center_of_mass_velocity()
        
        self.assertAlmostRelativeEquals(center_of_mass_before, center_of_mass_after, 8)
        self.assertAlmostRelativeEquals(center_of_mass_velocity_before, center_of_mass_velocity_after, 8)
        separation = (binary[0].position  - binary[1].position).length()
        self.assertAlmostRelativeEquals( separation, 2.5 | nbody_system.length)
    
    
    def test4(self):
        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.AU)
        kepler = Kepler(converter)
        kepler.initialize_code()
        x = encounters.KeplerOrbits(kepler)
        binary = new_binary( 
            1 | units.MSun,
            0.5 | units.MSun,
            1.2 | units.AU,
            G = constants.G
        )
        semimajor_axis, eccentricity = x.get_semimajor_axis_and_eccentricity_for_binary_components(
            binary[0],
            binary[1]
        )
        self.assertAlmostRelativeEquals(semimajor_axis,  1.2 | units.AU, 8)
        self.assertAlmostRelativeEquals(eccentricity,  0)

class TestScaleSystem(amusetest.TestWithMPI):
    
    def new_kepler(self, converter = None):
        x = Kepler(converter)
        x.initialize_code()
        return x
        
    def test1(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary( 
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
            0.5,
            is_at_periapsis = False
        )
        binary.radius = 0 |  nbody_system.length
        
        x = encounters.ScaleSystem(kepler)
        self.assertTrue((binary[0].position - binary[1].position).length() > 1.5 | nbody_system.length)
        x.scale_particles_to_sphere(binary, 0.75 | nbody_system.length)
        self.assertTrue((binary[0].position - binary[1].position).length() <= 1.6 | nbody_system.length)
        print((binary[0].position - binary[1].position).length())
        self.assertTrue((binary[0].position - binary[1].position).length() >= (1.5 - 1e-6)| nbody_system.length )
        
    
    
    def test2(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        binary = new_binary( 
            1 | nbody_system.mass,
            0.5 | nbody_system.mass,
            1.2 | nbody_system.length,
            0.5,
            is_at_periapsis = True
        )
        binary.radius = 0 |  nbody_system.length
        
        x = encounters.ScaleSystem(kepler)
        self.assertTrue((binary[0].position - binary[1].position).length() < 1.5 | nbody_system.length)
        x.scale_particles_to_sphere(binary, 0.75 | nbody_system.length)
        self.assertTrue((binary[0].position - binary[1].position).length() <= 1.6 | nbody_system.length)
        print((binary[0].position - binary[1].position).length())
        self.assertTrue((binary[0].position - binary[1].position).length() >= (1.5 - 1e-6)| nbody_system.length )
        
      
    def test3(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        
        particles= Particles(keys=(1,2))
        particles.mass = 1 | nbody_system.mass
        particles.child2 = None
        particles.child1 = None
        particles[0].position = [1,0,0] | nbody_system.length
        particles[0].velocity = [1,0.0,0] | nbody_system.speed
        
        particles[1].position = [-1,0,0] | nbody_system.length
        particles[1].velocity = [-1,0.0,0] | nbody_system.speed
        
        particles.radius = 0 |  nbody_system.length
        
        x = encounters.ScaleSystem(kepler)
        
        print((particles[0].position - particles[1].position).length())     
        self.assertTrue((particles[0].position - particles[1].position).length() > 1.5 | nbody_system.length)
        
        x.scale_particles_to_sphere(particles, 0.75 | nbody_system.length)
        
        self.assertTrue((particles[0].position - particles[1].position).length() <= 1.6 | nbody_system.length)
        print(particles)
        self.assertTrue((particles[0].position - particles[1].position).length() >= (1.5 - 1e-6)| nbody_system.length )
        self.assertAlmostRelativeEquals(particles[0].position,  [0.75,0,0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[1].position,  [-0.75,0,0] | nbody_system.length)
        
        
    def test4(self):
        kepler = encounters.KeplerOrbits(self.new_kepler())
        
        particles= Particles(keys=(1,2,3,4,5,6))
        particles.child2 = None
        particles.child1 = None
        particles.mass = 1 | nbody_system.mass
        for i in range(3):
            position = [0,0,0] | nbody_system.length
            position[i] = 1 | nbody_system.length
            particles[i].position =  position
            position[i] = -1 | nbody_system.length
            particles[i+3].position =  position
            
            velocity = [0,0,0] | nbody_system.speed
            velocity[i] = 1 | nbody_system.speed
            particles[i].velocity =  velocity
            velocity[i] = -1 | nbody_system.speed
            particles[i+3].velocity =  velocity

        particles.radius = 0 |  nbody_system.length
        
        
        x = encounters.ScaleSystem(kepler)
        potential_energy0 = particles.potential_energy(G = nbody_system.G)
        kinetic_energy0 = particles.kinetic_energy()
        
        x.scale_particles_to_sphere(particles, 0.5 | nbody_system.length)
        
        self.assertAlmostRelativeEquals(particles[0].position,  [1/numpy.sqrt(2),0,0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[3].position,  [-1/numpy.sqrt(2),0,0] | nbody_system.length)
        self.assertAlmostRelativeEquals(particles[0].velocity,  [1.542,0,0] | nbody_system.speed, 3)
        self.assertAlmostRelativeEquals(particles[3].velocity,  [-1.542,0,0] | nbody_system.speed, 3)
        
        potential_energy1 = particles.potential_energy(G = nbody_system.G)
        kinetic_energy1 = particles.kinetic_energy()
        
        
        self.assertAlmostRelativeEquals(potential_energy0 + kinetic_energy0,potential_energy1 + kinetic_energy1)

    def test5(self):
        converter = nbody_system.nbody_to_si(1 | units.MSun, 1 | units.AU)
        kepler = encounters.KeplerOrbits(self.new_kepler(converter))
        
        particles= Particles(keys=(1,2,3))
        particles.child2 = None
        particles.child1 = None
        particles.position = [
            [ -1.28230200e-05,  -3.69457095e-05,  -2.02383488e-05],
            [ -2.91749746e-05,  -1.21387289e-05,   1.56377986e-07],
            [  2.92123436e-05,   1.22463965e-05,  -9.73992061e-08]
        ] | units.parsec
        particles.velocity = [
            [  -8685.98414414, -413260.30543051, -135268.19175611],    
            [  19639.55455979,  -30251.55372943,     972.69648982],
            [ -19614.24178601,   31455.88066005,    -578.49669843]
        ] | (units.m / units.s)
        particles.mass = [0.14571045,  50., 50.] | units.MSun
        particles.radius = [8, 0., 0.0] | units.AU
        
        x = encounters.ScaleSystem(kepler, G = constants.G)
        
        potential_energy0 = particles.potential_energy()
        kinetic_energy0 = particles.kinetic_energy()
        
        x.scale_particles_to_sphere(particles, 4 | units.AU)
        
        potential_energy1 = particles.potential_energy()
        kinetic_energy1 = particles.kinetic_energy()
        
        self.assertAlmostRelativeEquals(potential_energy0 + kinetic_energy0,potential_energy1 + kinetic_energy1)
        
        
class TestHandleEncounter(amusetest.TestWithMPI):
    
    def new_kepler(self):
        x = Kepler()
        x.initialize_code()
        return x
        
    def test1(self):
        particles_in_encounter = Particles(keys=(1,2,3))
        particles_in_encounter.mass = 1. | nbody_system.mass
        particles_in_encounter[0].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter[2].position = [0.,0.5,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.0,0.] | nbody_system.speed
        particles_in_encounter[2].velocity = [0.,0.0,3.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length
        
        particles_in_field = Particles()
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(x.captured_singles.key, [1,2])
    

    def test2(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.0,0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length
        
        
        binary1 = new_binary(
            1. | nbody_system.mass, 
            1. | nbody_system.mass, 
            0.01 | nbody_system.length,
            keyoffset = 30
        )
        binary2 = new_binary(
            1. | nbody_system.mass, 
            1. | nbody_system.mass, 
            0.01 | nbody_system.length,
            keyoffset = 40
        )
        binaries = Particles(keys=(20,21))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binaries[1].child1 = binary2[0]
        binaries[1].child2 = binary2[1]
        binary1.child1 = None
        binary1.child2 = None
        binary2.child1 = None
        binary2.child2 = None
        
        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)
        multiple = particles_in_encounter[1]
        multiple.components = binary2
        multiples.add_particle(multiple)
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)
        
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 2)
        self.assertEqual(len(x.new_binaries), 2)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
    
    
        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, binaries[1].child1.key)
        multiple = x.new_multiples[1]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child2.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, binaries[1].child2.key)

    def test3(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.0,0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length
        
        
        binary1 = new_binary(
            1. | nbody_system.mass, 
            1. | nbody_system.mass,
            0.01 | nbody_system.length,
            keyoffset = 30
        )
        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None
        
        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)
        
        x.execute()
        
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.dissolved_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 1)
        self.assertEqual(len(x.released_singles), 1)
        
        multiple = x.new_multiples[0]
        print(multiple.child1)
        self.assertEqual(len(multiple.components), 2)
        print(multiple.components)
        
    def test4(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1.0 /20.0  | nbody_system.mass
        
        particles_in_encounter[0].position = [0.303976184547589401, 0.273137803329168094, 0.0] | nbody_system.length
        particles_in_encounter[1].position = [0.290167020486133631, 0.273139253307546515, 0.0] | nbody_system.length
        particles_in_encounter[0].velocity = [-2.544712989335638387, -1.224759650260411004, 0.0] | nbody_system.speed
        particles_in_encounter[1].velocity = [0.898326624897966997, 0.870611747842778838, 0.0] | nbody_system.speed
        particles_in_encounter.radius = 0.007 | nbody_system.length
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        x.execute()
        
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        r_before = (particles_in_encounter[0].position - particles_in_encounter[1].position ).length()
        r_after = (x.particles_after_encounter[0].position - x.particles_after_encounter[1].position).length()
        print(r_before, r_after)
        self.assertFalse(r_after > (10 * r_before))
        
    def test5(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1.0 /20.0  | nbody_system.mass
        
        particles_in_encounter[0].position = [0.0779377282404, -0.559210143918, 0.0] | nbody_system.length
        particles_in_encounter[1].position = [0.07802860386, -0.561207706614, 0.0]  | nbody_system.length
        particles_in_encounter[0].velocity = [-1.79086847491, -4.88551917358, 0.0] | nbody_system.speed
        particles_in_encounter[1].velocity = [2.13208424698, 4.31403500143, 0.0] | nbody_system.speed
        particles_in_encounter.radius = 0.001 | nbody_system.length
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        x.execute()
        
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        r_before = (particles_in_encounter[0].position - particles_in_encounter[1].position ).length()
        r_after = (x.particles_after_encounter[0].position - x.particles_after_encounter[1].position).length()
        print(r_before, r_after)
        self.assertFalse(r_after > (10 * r_before))
            
    
    def test6(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 2 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles(10)
        particles_in_field.mass = 1 | nbody_system.mass
        for i in range(len(particles_in_field)):
            if i == 5:
                j = len(particles_in_field)
            else:
                j = i
            particles_in_field[i].position = [-10 + (2*j), -10 + (2*j), 0] | nbody_system.length
        particles_in_field.velocity =  [0,0.0,0] | nbody_system.speed
        particles_in_field.radius = 0 | nbody_system.length
        
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.particles_in_field.add_particles(particles_in_field)
        
        x.execute()
        
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.dissolved_multiples), 0)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(len(x.released_singles), 0)
        
    
    def test7(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 2. | nbody_system.mass
        particles_in_encounter[0].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.0,0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length
        
        
        binary1 = new_binary(
            1. | nbody_system.mass, 
            1. | nbody_system.mass, 
            0.01 | nbody_system.length,
            keyoffset = 30
        )
        binary_energy = binary1.kinetic_energy() + binary1.potential_energy(G = nbody_system.G)
        self.assertAlmostRelativeEqual(binary_energy, -50 | nbody_system.energy)
        
        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None
        
        
        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)
        
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        self.assertEqual(len(x.captured_singles), 1)
        self.assertEqual(len(x.released_singles), 1)
        
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
    
        self.assertAlmostRelativeEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy, -50 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy, -85.651753404 | nbody_system.energy, 4)
        
    
    def test8(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter[0].mass = 2. | nbody_system.mass
        particles_in_encounter[1].mass = 0.1 | nbody_system.mass
        particles_in_encounter[0].position = [1.,0.,0.] | nbody_system.length
        particles_in_encounter[1].position = [0.,0.,0.] | nbody_system.length
        particles_in_encounter.velocity = [0.,0.0,0.] | nbody_system.speed
        particles_in_encounter.radius = 0. | nbody_system.length
        
        
        binary1 = new_binary(
            1. | nbody_system.mass, 
            1. | nbody_system.mass, 
            0.01 | nbody_system.length,
            keyoffset = 30
        )

        binary_energy = binary1.kinetic_energy() + binary1.potential_energy(G = nbody_system.G)
        self.assertAlmostRelativeEqual(binary_energy, -50 | nbody_system.energy)
        
        binaries = Particles(keys=(20,))
        binaries[0].child1 = binary1[0]
        binaries[0].child2 = binary1[1]
        binary1.child1 = None
        binary1.child2 = None
        
        
        multiples = Particles()
        multiple = particles_in_encounter[0]
        multiple.components = binary1
        multiples.add_particle(multiple)
        
        x = encounters.HandleEncounter(
            kepler_code = self.new_kepler(),
            resolve_collision_code = SmallN(),
            interaction_over_code = None,
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        x.existing_binaries.add_particles(binaries)
        x.existing_multiples.add_particles(multiples)
        
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 0)
        self.assertEqual(len(x.updated_multiples), 1)
        self.assertEqual(len(x.new_binaries), 0)
        self.assertEqual(len(x.captured_singles), 0)
        self.assertEqual(len(x.released_singles), 0)
        
        multiple = x.updated_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertEqual(multiple.key, multiples[0].key)
    
        self.assertEqual(multiple.components[0].key, binaries[0].child1.key)
        self.assertEqual(multiple.components[1].key, binaries[0].child2.key)
        self.assertAlmostRelativeEqual(x.initial_multiple_energy, -50 | nbody_system.energy)
        self.assertAlmostRelativeEqual(x.final_multiple_energy,  -56.2906032319 | nbody_system.energy, 4)
        
        
class TestStickyHandleEncounter(amusetest.TestWithMPI):

    def test1(self):
        particles_in_encounter = Particles(keys=(1,2))
        particles_in_encounter.mass = 1 | nbody_system.mass
        particles_in_encounter[0].position = [1,0,0] | nbody_system.length
        particles_in_encounter[1].position = [0,0,0] | nbody_system.length
        particles_in_encounter.velocity = [0,0.0,0] | nbody_system.speed
        particles_in_encounter[0].velocity = [-1,0.0,0] | nbody_system.speed
        particles_in_encounter.radius = 0 | nbody_system.length
        
        particles_in_field = Particles()
        
        x = encounters.StickyHandleEncounter(
            G = nbody_system.G
        )
        
        x.particles_in_encounter.add_particles(particles_in_encounter)
        
        x.execute()
        self.assertEqual(len(x.new_multiples), 1)
        self.assertEqual(len(x.new_binaries), 1)
        multiple = x.new_multiples[0]
        self.assertEqual(len(multiple.components), 2)
        self.assertAlmostRelativeEqual(multiple.components[0].key, particles_in_encounter[0].key)
        self.assertAlmostRelativeEqual(multiple.components[1].key, particles_in_encounter[1].key)
        self.assertEqual(len(x.captured_singles), 2)
        self.assertEqual(x.captured_singles.key, [1,2])
        self.assertAlmostRelativeEqual(multiple.velocity, [-0.5,0.0,0] | nbody_system.speed)
        self.assertAlmostRelativeEqual(multiple.mass, 2 |  nbody_system.mass)
        self.assertAlmostRelativeEqual(multiple.radius, 0.5 |  nbody_system.length)
        self.assertAlmostRelativeEqual(multiple.position, [0.5,0,0] | nbody_system.length)
        
        
        



from amuse.test import amusetest

from amuse.community.sse.interface import SSE
from amuse.community.bhtree.interface import BHTree

class TestsForIssue123(amusetest.TestCase):
  def test1(self): # doesn't trigger recursion error
    self.assertRaises(Exception, BHTree, name_of_the_worker="bogus",expected_message= 
      "__init__() got multiple values for keyword argument 'name_of_the_worker'")
  def test2(self): # does
    self.assertRaises(Exception, SSE, name_of_the_worker="bogus",expected_message= 
      "__init__() got multiple values for keyword argument 'name_of_the_worker'")



from amuse.test import amusetest

import os
import numpy
import time

from amuse import io
from amuse.io import store
from amuse.io import store_v1
from amuse.io import store_v2
from amuse.units import units
from amuse.units import nbody_system
from amuse.datamodel import Particles
from amuse.datamodel import ParticlesOverlay
from amuse.datamodel import Grid
from amuse.datamodel import ParticlesSuperset




class _AbstractTestStoreHDF(object):
    
    def store_factory(self):
        raise NotImplementedError
    
    def store_version(self):
        raise NotImplementedError
        
    def get_version_in_store(self, container):
        raise NotImplementedError

    def test1(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test1."+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        instance.store(p)

        loaded_particles = instance.load()

        loaded_mass_in_kg = loaded_particles.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)
        
        instance.close()

    def test2(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test2."+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        instance.store(p.savepoint(1 | units.Myr))

        p.mass = [x * 4.0 for x in range(number_of_particles)] | units.kg
        instance.store(p.savepoint(2 | units.Myr))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(self.get_version_in_store(loaded_particles).get_timestamp(), 2 | units.Myr)
        self.assertAlmostRelativeEquals(self.get_version_in_store(loaded_particles).previous_state().get_timestamp(), 1 | units.Myr)
        masses = loaded_particles[1].get_timeline_of_attribute("mass")
        self.assertEqual(len(masses), 2)

        instance.close()

    def test3(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test3"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | nbody_system.mass

        instance.store(p.savepoint(1 | nbody_system.time))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(p.mass[1], 2.0 | nbody_system.mass)

        instance.close()

    def test4(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test4"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(10)
        particles.mass = 1.0 | units.kg
        particles.x = 2.0 | units.m
        x = particles.savepoint(2.0 | units.s)

        io.write_set_to_file(x,output_file, format='amuse', version = self.store_version())

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version = self.store_version(), close_file = True)
        particles_in_memory = particles_from_file.copy()

        self.assertAlmostRelativeEquals(particles_in_memory.mass[1], 1.0 | units.kg)

        particles_in_memory.savepoint(4.0 | units.s)

        self.assertAlmostRelativeEquals(particles_from_file.mass[2], 1.0 | units.kg)
    
    def test5(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test5_grid"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)
        
        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        instance.store_grid(p)

        loaded_grid = instance.load_grid()

        self.assertEqual(loaded_grid.shape, shape)
        
        loaded_mass_in_kg = loaded_grid.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)
        
        instance.close()
    
    def test6(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test6_grid"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)
        
        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s
        
        instance.store_grid(p.savepoint(1| units.Myr))

        loaded_grid = self.get_version_in_store(instance.load_grid())
        
        self.assertAlmostRelativeEquals(loaded_grid.get_timestamp(), 1| units.Myr)
        
        self.assertEqual(loaded_grid.shape, shape)
        self.assertEqual(loaded_grid[0][0][0].mass, 0 | units.kg)
        self.assertAlmostRelativeEquals(loaded_grid[...,1,1].mass, p[...,1,1].mass)

        instance.close()
        
    def test7(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test7"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        instance.store(p)
        instance.close()
    
        instance = self.store_factory()(output_file)
        loaded_particles =  self.get_version_in_store(instance.load())

        loaded_particles.mass = [x * 3.0 for x in range(number_of_particles)] | units.kg
        previous_mass_in_kg = [x * 3.0 for x in range(number_of_particles)]
        instance.close()
        
        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        loaded_mass_in_kg = loaded_particles.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)
        
        instance.close()
        
        instance = self.store_factory()(output_file)
        loaded_particles = self.get_version_in_store(instance.load())
        loaded_particles[2].mass = 44 | units.kg
        instance.close()
        instance = self.store_factory()(output_file)
        loaded_particles = self.get_version_in_store(instance.load())
        self.assertEqual( 44 | units.kg, loaded_particles[2].mass)
        instance.close()
        
    def test8(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test8"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        instance = self.store_factory()(output_file)
        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = numpy.asarray([x * 2.0 for x in range(number_of_particles)])

        instance.store(p.savepoint(1 | nbody_system.time))
        instance.close()

        instance = self.store_factory()(output_file)
        loaded_particles = instance.load()
        self.assertAlmostRelativeEquals(p.mass[1], 2.0)
        instance.close()
    

    def test9(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test9"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        io.write_set_to_file(
            p,
            output_file,
            "hdf5", 
            timestamp = 2 | units.Myr,
            scale = 1 | units.kg,
            version = self.store_version(),
            close_file = True
        )
        
        loaded_particles = io.read_set_from_file(
            output_file, 
            "hdf5",
            version = self.store_version()
        )
        loaded_particles = self.get_version_in_store(loaded_particles)
        
        
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test10(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test10"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p, output_file, "hdf5",  timestamp = 2 | units.Myr, scale = 1 | units.kg, version = self.store_version())
        
        loaded_particles = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)

    def test11(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test11"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p.savepoint(timestamp = 2 | units.Myr, scale = 1 | units.kg), output_file, "hdf5", version = self.store_version())
        
        loaded_particles = io.read_set_from_file(output_file, "hdf5", version = self.store_version())        
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)
        
    
    def test12(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test12"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        io.write_set_to_file(p.savepoint(timestamp = 2 | units.Myr, scale = 1 | units.kg), output_file, "hdf5", version = self.store_version())
        
        loaded_particles = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        loaded_particles = self.get_version_in_store(loaded_particles)
        a = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(a.timestamp, 2 | units.Myr)
        self.assertAlmostRelativeEquals(a.scale, 1 | units.kg)
    
    def test13(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test13"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0  | units.km
        stars.md = [[1,2,3],[4,5,6]] | units.km
       

        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        self.assertEqual(loaded[0].md, [1,2,3] | units.km)
        self.assertEqual(loaded[1].md, [4,5,6] | units.km)
        
        self.assertEqual(loaded.md[0], [1,2,3] | units.km)
        self.assertEqual(loaded.md[1], [4,5,6] | units.km)
         
        self.assertEqual(self.get_version_in_store(loaded)[0].md, [1,2,3] | units.km)
        #self.assertEquals(self.get_version_in_store(loaded)[0].md,  [7,8,9] | units.km)
    
      
    def test14(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test14"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0  | units.km
        stars.md = [[[1,3],[2,4],[3,5]],[[4,6],[5,7],[6,8]]]
       

        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        self.assertEqual(loaded[0].md, [[1,3],[2,4],[3,5]])
        self.assertEqual(loaded[1].md, [[4,6],[5,7],[6,8]])
        
        self.assertEqual(loaded.md[0], [[1,3],[2,4],[3,5]])
        self.assertEqual(loaded.md[1], [[4,6],[5,7],[6,8]])
    
    def test15(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test15"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0  | units.km
        stars.md = [[[1,3],[2,4],[3,5]],[[4,6],[5,7],[6,8]]]
       

        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        processor = self.store_factory()(output_file, True, open_for_writing = True)
        loaded = processor.load()
        self.get_version_in_store(loaded)[0].md = [[3,1],[3,4],[5,2]] 
        self.assertEqual(self.get_version_in_store(loaded)[0].md, [[3,1],[3,4],[5,2]] )
        processor.close()
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        self.assertEqual(loaded[0].md, [[3,1],[3,4],[5,2]])
        self.assertEqual(loaded[1].md, [[4,6],[5,7],[6,8]])
        
    def test16(self):
        import h5py
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test16"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        stars[0].nn = stars[1]
        stars[1].nn = stars[0]
       

        self.assertEqual(stars[0].nn,stars[1])
        self.assertEqual(stars[1].nn,stars[0])
        
        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        processor = self.store_factory()(output_file, True, open_for_writing = True)
        loaded = processor.load()
        
        self.assertEqual(loaded[0].nn,loaded[1])
        self.assertEqual(loaded[1].nn,loaded[0])
        self.assertEqual(loaded[0].nn.key,stars[1].key)
        self.assertEqual(loaded[0].nn.x,stars[1].x)
        
    
    def test17(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test17"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(4)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        stars[2].x = 3.0  | units.km
        stars[3].x = 4.0  | units.km
        
        binaries = Particles(2)
        binaries[0].y = 1.0 | units.km
        binaries[1].y = 2.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]
        binaries[1].child1 = stars[2]
        binaries[1].child2 = stars[3]

        self.assertEqual(binaries[0].child1,stars[0])
        self.assertEqual(binaries[1].child1,stars[2])
        
        io.write_set_to_file(binaries, output_file, "hdf5", version = self.store_version())
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertEqual(loaded[0].child1.key,stars[0].key)
        self.assertEqual(loaded[1].child1.key,stars[2].key)
        
    def test18(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test18"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]
        
        self.assertEqual(binaries[0].child1,stars[0])
        self.assertEqual(binaries[0].child2,stars[1])
        self.assertEqual(binaries[0].child1.parent,binaries[0])
        self.assertEqual(binaries[0].child2.parent,binaries[0])
        
        io.write_set_to_file(binaries, output_file, "hdf5", version = self.store_version())
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertEqual(loaded[0].child1.key,stars[0].key)
        self.assertEqual(loaded[0].child2.key,stars[1].key)
        self.assertEqual(loaded[0].child1,stars[0])
        self.assertEqual(loaded[0].child2,stars[1])
        self.assertEqual(loaded[0].child1.parent,loaded[0])
        self.assertEqual(loaded[0].child2.parent,loaded[0])
    
    def test19(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test19"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]
        
        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]
        
        self.assertEqual(binaries[0].child1,stars[0])
        self.assertEqual(binaries[0].child2,stars[1])
        self.assertEqual(binaries[0].child1.parent,binaries[0])
        self.assertEqual(binaries[0].child2.parent,binaries[0])
        
        io.write_set_to_file([binaries,stars], output_file, "hdf5", names = ['binaries', 'children'], version = self.store_version())
        
        loader_binaries, loaded_stars = io.read_set_from_file(output_file, "hdf5", names = ['binaries', 'children'], version = self.store_version())
        
        self.assertEqual(loader_binaries[0].child1.key,stars[0].key)
        self.assertEqual(loader_binaries[0].child2.key,stars[1].key)
        self.assertEqual(loader_binaries[0].child1,loaded_stars[0])
        self.assertEqual(loader_binaries[0].child2,loaded_stars[1])
        self.assertEqual(loader_binaries[0].child1.parent,loader_binaries[0])
        self.assertEqual(loader_binaries[0].child2.parent,loader_binaries[0])

    def test20(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test20"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        shape = 10, 10, 10
        p = Grid(*shape)
        p.mass = (numpy.asarray([x * 2.0 for x in range(p.size)])).reshape(shape)

        io.write_set_to_file(p, output_file, "hdf5", version = self.store_version())
        
        loaded = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        loaded = self.get_version_in_store(loaded)
         
        self.assertAlmostRelativeEquals(p.mass[0][1][2], 24)
        self.assertAlmostRelativeEquals(p[0][1][2].mass, 24)
        
    def test21(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test21"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0  | units.km
        stars.md = [[[1,3],[2,4],[3,5]],[[4,6],[5,7],[6,8]]]
       

        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
       
        loaded = io.read_set_from_file(output_file, "hdf5", close_file = True, version = self.store_version(), copy_history=False)
        
        self.assertEqual(loaded[0].md, [[1,3],[2,4],[3,5]])
        self.assertEqual(loaded[1].md, [[4,6],[5,7],[6,8]])
        
        previous = loaded.previous_state()
        self.assertEqual(previous, None)


    def test22(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test22"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars.x = 1.0  | units.km
       
        overlay = ParticlesOverlay(stars)
        overlay.y = 2.0  | units.km


        io.write_set_to_file(overlay, output_file, "hdf5", version = self.store_version())
       
        loaded = io.read_set_from_file(output_file, "hdf5", close_file = True, version = self.store_version())
        
        self.assertEqual(loaded[0].x,  1.0 | units.km)
        self.assertEqual(loaded[1].y,  2.0 | units.km)
        
    
    def test23(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test23"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles()

        io.write_set_to_file(particles, output_file, format='amuse', version = self.store_version())

        particles_from_file = io.read_set_from_file(output_file, format='amuse', version = self.store_version())
        
        self.assertEqual(len(particles_from_file), 0)
    
    def test24(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test24"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(20)
#~        particles = Particles(1000000) # For testing memory usage
        particles.mass = 1.0 | units.kg
        particles.position = [0.0, 0.0, 0.0] | units.m
        
        for i in range(10):
            particles.position += [1.0, 2.0, 3.0] | units.m
            io.write_set_to_file(particles, output_file, format='amuse', version = self.store_version(), append_to_file=True)
        
        particles_from_file = io.read_set_from_file(output_file, format='amuse', version = self.store_version())
        os.remove(output_file)
        self.assertEqual(len(list(particles_from_file.history)), 10)
        for i, snap in enumerate(particles_from_file.history):
            self.assertEqual(len(snap), 20)
            self.assertEqual((i+1) * ([1.0, 2.0, 3.0] | units.m), snap.center_of_mass())
            
    def test25(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test25"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(20)
#~        particles = Particles(1000000) # For testing memory usage
        particles.mass = 1.0 | units.kg
        particles.position = [0.0, 0.0, 0.0] | units.m
        
        for i in range(10):
            particles.position += [1.0, 2.0, 3.0] | units.m
            io.write_set_to_file(particles, output_file, format='amuse', version = self.store_version(), append_to_file=True)
        
        particles_from_file = io.read_set_from_file(output_file, format='amuse', version = self.store_version(), copy_history = True, close_file = True)
        
        history = list(particles_from_file.history)
        self.assertEqual(len(history), 10)
        self.assertFalse("HDF" in str(type(history[1]._private.attribute_storage)))
        for i, snap in enumerate(particles_from_file.history):
            self.assertEqual(len(snap), 20)
            self.assertEqual((i+1) * ([1.0, 2.0, 3.0] | units.m), snap.center_of_mass())
        os.remove(output_file)
        
    def test26(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        p = Grid(20)
        
        io.write_set_to_file(p, output_file, format='amuse', version = self.store_version())
        
        os.remove(output_file)

    def test27(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test27"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)
        instance = self.store_factory()(output_file)
        
        shape = 10,
        p = Grid(*shape)
        p.mass = ([x * 2.0 for x in range(p.size)] | units.kg).reshape(shape)
        p.model_time = 2.0 | units.s

        instance.store_grid(p)

        loaded_grid = instance.load_grid()

        self.assertEqual(loaded_grid.shape, shape)
        
        loaded_mass_in_kg = loaded_grid.mass.value_in(units.kg)
        previous_mass_in_kg = p.mass.value_in(units.kg)
        for expected, actual in zip(previous_mass_in_kg, loaded_mass_in_kg):
            self.assertEqual(expected, actual)
        
        instance.close()

    def test28(self):
        
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test28"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        
        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        binaries = Particles(1)
        binaries[0].y = 1.0 | units.km
        binaries[0].child1 = stars[0]
        binaries[0].child2 = stars[1]

        stars[0].parent = binaries[0]
        stars[1].parent = binaries[0]
        
        self.assertEqual(binaries[0].child1,stars[0])
        self.assertEqual(binaries[0].child2,stars[1])
        self.assertEqual(binaries[0].child1.parent,binaries[0])
        self.assertEqual(binaries[0].child2.parent,binaries[0])
        
        io.write_set_to_file(binaries, output_file, "hdf5", version = self.store_version())
        
        with io.read_set_from_file(output_file, "hdf5", version = self.store_version(), return_context = True) as loaded:
        
            self.assertEqual(loaded[0].child1.key,stars[0].key)
            self.assertEqual(loaded[0].child2.key,stars[1].key)
            self.assertEqual(loaded[0].child1,stars[0])
            self.assertEqual(loaded[0].child2,stars[1])
            self.assertEqual(loaded[0].child1.parent,loaded[0])
            self.assertEqual(loaded[0].child2.parent,loaded[0])

    def test29(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test29"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Grid(10)
        particles.attribute1 = "normal"
        particles.attribute2 = "unicode"

        io.write_set_to_file(particles,output_file, format='amuse', version = self.store_version())
        
        output = io.read_set_from_file(output_file, format='amuse')
        self.assertEqual(output[0].attribute2, "unicode")
        
    def test30(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test30"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles = Particles(10)
        particles.mass = 10 | units.kg

        io.write_set_to_file(particles,output_file, format='amuse', version = self.store_version())
        
        output = io.read_set_from_file(output_file, format='amuse', allow_writing = True)
        output.new_attribute= 2 * output.mass

        self.assertEqual(output.new_attribute, 2 * output.mass)
        output = next(output.iter_history())
        output.new_attribute= 2 * output.mass

        self.assertEqual(output.new_attribute, 2 * output.mass)
        
    def test31(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test31"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        p=Particles(10)
        p.mass=numpy.arange(10)
        q=Particles(3)
        q[0].link=p[1]
        q[1].link=p[3]
        q[2].link=p[4]
        io.write_set_to_file(q,output_file,"amuse",version=self.store_version())
        r=io.read_set_from_file(output_file,"amuse")
        self.assertEqual(len(r.link.shape),1)
        self.assertEqual(r[0].link.mass,1)
        self.assertEqual(r[1].link.mass,3)
        self.assertEqual(r[2].link.mass,4)



class TestStoreHDFV1(amusetest.TestCase, _AbstractTestStoreHDF):
    
    def store_factory(self):
        return store_v1.StoreHDF
    
    def store_version(self):
        return '1.0'
        
    def get_version_in_store(self, container):
        return container.previous_state()
    
class TestStoreHDFV2(amusetest.TestCase, _AbstractTestStoreHDF):
    
    def store_factory(self):
        return store_v2.StoreHDF
    
    def store_version(self):
        return '2.0'
        
    def get_version_in_store(self, container):
        return container
        
    def test52(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test52"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        gas = Particles(1)
        gas[0].y = 1.0 | units.km
        
        stars[0].gas = gas
        
        
        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        
        loaded_stars = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertEqual(loaded_stars[0].gas[0].key,gas[0].key)
        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0].y,1.0 | units.km)
        self.assertEqual(loaded_stars[1].gas, None)

        
    def test53(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test53"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        gas = Grid(2,3)
        gas.y = [[1.0, 2.0, 3.0], [4.0,5.0,6.0]] | units.km
        
        stars[0].gas = gas
        
        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        loaded_stars = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0][0].y,1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_stars[0].gas[0][2].y,3.0 | units.km)
        self.assertEqual(loaded_stars[1].gas, None)

    def test54(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test23"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        gas = Grid(2,3)
        gas.y = [[1.0, 2.0, 3.0], [4.0,5.0,6.0]] | units.km
        
        stars[0].gas = gas[0][0]
        stars[1].gas = gas[1][0]
        
        self.assertAlmostRelativeEquals(stars[0].gas.y,1.0 | units.km)
        self.assertAlmostRelativeEquals(stars[1].gas.y,4.0 | units.km)
        
        io.write_set_to_file(stars, output_file, "hdf5", version = self.store_version())
        loaded_stars = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertAlmostRelativeEquals(loaded_stars[0].gas.y,1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_stars[1].gas.y,4.0 | units.km)

    def test54b(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test24"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        stars = Particles(2)
        stars[0].x = 1.0  | units.km
        stars[1].x = 2.0  | units.km
        
        gas = Grid(2,3)
        gas.y = [[1.0, 2.0, 3.0], [4.0,5.0,6.0]] | units.km
        gas[0][0].particle = stars[0]
        gas[1][2].particle = stars[0]
        gas[1][0].particle = stars[1]
                
        #stars[0].gas = gas[0][0]
        #stars[1].gas = gas[1][0]
        
        self.assertAlmostRelativeEquals(gas[0][0].particle.x,1.0 | units.km)
        self.assertAlmostRelativeEquals(gas[1][2].particle.x,1.0 | units.km)
        self.assertAlmostRelativeEquals(gas[1][0].particle.x,2.0 | units.km)
        
        io.write_set_to_file(gas, output_file, "hdf5", version = self.store_version())
        loaded_gas = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertAlmostRelativeEquals(loaded_gas[0][0].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_gas[1][2].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_gas[1][0].particle.x, 2.0 | units.km)
        self.assertEqual(id(loaded_gas[0][0].particle.get_containing_set()), id(loaded_gas[1][2].particle.get_containing_set()))
        
        self.assertEqual(loaded_gas.particle.shape, loaded_gas.shape)
                
        gas_copy = loaded_gas.copy()

        gas_copy[0][0].particle.x = 3 | units.km
        self.assertAlmostRelativeEquals(loaded_gas[0][0].particle.x, 1.0 | units.km)
        self.assertAlmostRelativeEquals(gas_copy[0][0].particle.x, 3.0 | units.km)
        self.assertAlmostRelativeEquals(gas_copy[1][2].particle.x, 3.0 | units.km)
        
    
    def test55(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test25"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        particles1 = Particles(2)
        particles1[0].x = 1.0  | units.km
        particles1[1].x = 2.0  | units.km
        particles2 = Particles(2)
        particles2[0].x = 3.0  | units.km
        particles2[1].x = 4.0  | units.km
        
        particles_superset = ParticlesSuperset([particles1, particles2])
    
        
        self.assertAlmostRelativeEquals(particles_superset[0].x,1.0 | units.km)
        self.assertAlmostRelativeEquals(particles_superset[2].x,3.0 | units.km)
        
        io.write_set_to_file(particles_superset, output_file, "hdf5", version = self.store_version())
        loaded_particles_superset = io.read_set_from_file(output_file, "hdf5", version = self.store_version())
        
        self.assertAlmostRelativeEquals(loaded_particles_superset[0].x,1.0 | units.km)
        self.assertAlmostRelativeEquals(loaded_particles_superset[2].x,3.0 | units.km)
        
    

    def test56(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        number_of_particles = 10
        p = Particles(number_of_particles)
        p.mass = [x * 2.0 for x in range(number_of_particles)] | units.kg
        p.model_time = 2.0 | units.s

        gas = Grid(2,3)
        gas.y = [[1.0, 2.0, 3.0], [4.0,5.0,6.0]] | units.km
        
        io.write_set_to_file(
            p,
            output_file,
            "hdf5", 
            particle = p[1],
            particles = p,
            gridpoint = gas[0][0],
            grid = gas,
            version = self.store_version()
        )
        
        loaded_particles = io.read_set_from_file(
            output_file, 
            "hdf5",
            version = self.store_version(), copy_history=False, close_file=False
        )
        
        attributes = loaded_particles.collection_attributes
        self.assertAlmostRelativeEquals(attributes.particle.mass, loaded_particles[1].mass)
        self.assertEquals(attributes.particle.key, loaded_particles[1].key)
        self.assertEquals(id(attributes.particles), id(loaded_particles))
        self.assertAlmostRelativeEquals(attributes.gridpoint.y, 1.0 | units.km)
        self.assertAlmostRelativeEquals(attributes.grid[0][0].y, 1.0 | units.km)

    def test57(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test26"+self.store_version()+".hdf5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(2)
        x.sub = None
        x[0].sub = Particles(2)
        io.write_set_to_file(x, output_file,"amuse", version=self.store_version())
        y = io.read_set_from_file(output_file,"amuse")
        
        self.assertEqual(len(x), len(y))
        self.assertEqual(len(x[0].sub), len(y[0].sub))
        self.assertTrue(y[1].sub==x[1].sub==None)

        os.remove(output_file)
    def test58(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test58"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        x = Particles(keys = (1,2))
        x.mass = [1,2] | units.kg
        y = Particles(keys = (11,12,13,14))
        y.mass = [40,50,60,70] | units.kg
        
        x.sub = None
        x[0].sub = y[0:2]
        y[0].sub = y[2:]
        
        io.write_set_to_file(x, output_file,"amuse", version=self.store_version())
        z = io.read_set_from_file(output_file,"amuse")
        
        self.assertEqual(len(x), len(z))
        self.assertEqual(len(x[0].sub), len(z[0].sub))
        self.assertEqual(x[0].sub[0].sub.key, (13,14))
        self.assertEqual(z[0].sub[0].sub.key, (13,14))
        self.assertEqual(id(x[0].sub._original_set()), id(x[0].sub[0].sub._original_set()))
        # no more subsets, could this be fixed, would be very nice to do so
        # self.assertEqual(id(z[0].sub._original_set()), id(z[0].sub[0].sub._original_set()))
        self.assertTrue(z[1].sub==x[1].sub==None)

        os.remove(output_file)

    def test59(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test59"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

        g=Grid(10,10)

        p=Particles(2)

        p[0].sub=g[0:5]

        io.write_set_to_file(p, output_file,"amuse", version=self.store_version())
        z = io.read_set_from_file(output_file,"amuse")
        
        os.remove(output_file)
    def test60(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test60"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        x = Particles(2)
        x.mass = 1 | units.kg
        y = Particles(2)
        y.mass = 2 | units.kg
        io.write_set_to_file((x,y), output_file,"amuse", version=self.store_version(), names = ("x", "y"))
        z = io.read_set_from_file(output_file,"amuse")
        
        os.remove(output_file)
        self.assertTrue("x" in z)
        self.assertTrue("y" in z)
        self.assertAlmostRelativeEquals(z["x"].mass, 1 | units.kg)
        self.assertAlmostRelativeEquals(z["y"].mass, 2 | units.kg)


    def test61(self):
        test_results_path = self.get_path_to_results()
        output_file = os.path.join(test_results_path, "test61"+self.store_version()+".h5")
        if os.path.exists(output_file):
            os.remove(output_file)

       
        x = Particles(10)
        x.mass = 1 | units.kg
        y = Particles(20)
        y.id = list(range(20))
        x[0].y = y[5:7]

        io.write_set_to_file(x, output_file,"amuse", version=self.store_version())
        z = io.read_set_from_file(output_file,"amuse")
        self.assertEqual(x[0].y[0].id, 5)
        self.assertEqual(z[0].y[0].id, 5)

        os.remove(output_file)

        
        



from amuse.test import amusetest
from io import StringIO
import textwrap
import os
import numpy

from amuse import io
from amuse.io import text
from amuse.units import units
from amuse.units import quantities, core
from amuse.units import generic_unit_system
from amuse import datamodel

class CursorTests(amusetest.TestCase):
    
    def test1(self):
        contents = "1\n2\n3"
        data_file = StringIO(contents)
        instance = text.LineBasedFileCursor(data_file)
        
        self.assertEqual("1", instance.line())
        self.assertEqual("1", instance.line())
        instance.forward()
        self.assertEqual("2", instance.line())
        instance.forward()
        self.assertEqual("3", instance.line())
        self.assertEqual("3", instance.line())
    
    def test2(self):
        contents = "1\n2\n3"
        data_file = StringIO(contents)
        instance = text.LineBasedFileCursor(data_file)
        
        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertTrue(instance.is_at_end())
        
class TableFormattedTextTests(amusetest.TestCase):
    
    def test1(self):
        contents = "#header\n1 2 3\n4 5 6\n       \n7 8 9\n     "
        data_file = StringIO(contents)
        instance = text.TableFormattedText("test.txt", data_file)
        instance.attribute_names = ['a', 'b', 'c']
        particles = instance.load()
        
        self.assertEqual(len(particles), 3)
        self.assertEqual(particles[0].a, 1)
        self.assertFalse(quantities.is_quantity(particles[0].a))
        
    def test2(self):
        p = datamodel.Particles(2)
        p.a = [1, 4] | units.m
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m
        
        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.attribute_types = [units.m, units.m, units.m]
        instance.store()
        
        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#m m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)
    
    def test3(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.MSun
        x.radius = [3.0, 4.0] | units.RSun
        io.write_set_to_file(x, "test.csv","txt", attribute_types = (units.MSun, units.RSun))
        with open("test.csv", "r") as f:
            contents = f.read()
        self.assertEqual("#mass radius\n#MSun RSun\n1.0 3.0\n2.0 4.0\n", contents)
    
        os.remove("test.csv")
        

    def test4(self):
        mass = [1.0,2.0,3.0] | generic_unit_system.mass
        length = [3.0,4.0,5.0] | generic_unit_system.length
        
        stream = StringIO()
        output = text.TableFormattedText(stream = stream)
        output.quantities = (mass, length)
        output.store()
        contents = stream.getvalue()
        self.assertEqual("#col(0) col(1)\n#mass length\n1.0 3.0\n2.0 4.0\n3.0 5.0\n", contents)
    
    def test5(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.MSun
        x.radius = [3.0, 4.0] | units.RSun
        
        expected = [
            "#mass radius\n#MSun RSun\n{0} 1.0 3.0\n{1} 2.0 4.0\n".format(x[0].key, x[1].key),
            "#mass radius\n#MSun RSun\n1.0 {0} 3.0\n2.0 {1} 4.0\n".format(x[0].key, x[1].key),
            "#mass radius\n#MSun RSun\n1.0 3.0 {0}\n2.0 4.0 {1}\n".format(x[0].key, x[1].key),
        ]
        for column_index, expected_content in enumerate(expected):
            io.write_set_to_file(
                x, 
                "test.csv",
                "txt", 
                key_in_column = column_index, 
                attribute_types = (units.MSun, units.RSun)
            )
                
            with open("test.csv", "r") as f:
                contents = f.read()
            
            self.assertEqual(expected_content, contents)
            
            y = io.read_set_from_file(
                "test.csv",
                "txt", 
                key_in_column = column_index, 
                attribute_types = (units.MSun, units.RSun),
                attribute_names = ('mass', 'radius')
            )
            
            self.assertEqual(y[0], x[0])
            self.assertEqual(y[1], x[1])
        os.remove("test.csv")
        
    
    
    def test6(self):
        p = datamodel.Particles(2)
        p.a = [1., 4.]
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m
        
        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.attribute_types = [None, units.m, units.m]
        instance.store()
        
        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#- m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)
        
    def test7(self):
        p = datamodel.Particles(2)
        p.a = [1.0, 4.0]
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m
        
        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.store()
        
        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#- m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)
        
    def test8(self):
        table = io.ReportTable(
            "test.csv",
            "txt", 
            attribute_types = (units.MSun, units.RSun)
        )
        table.add_row(1.0 | units.MSun, 3.0 | units.RSun)
        table.add_row(2.0 | units.MSun, 4.0 | units.RSun)
        table.close()
        
        with open("test.csv", "r") as f:
            contents = f.read()
        self.assertEqual("#MSun RSun\n1.0 3.0\n2.0 4.0\n", contents)
        
        
        os.remove("test.csv")
        
    def test9(self):
        p = datamodel.Particles(5)
        p.a = [1.0, 2.0, 3.0, 4.0, 5.0]
        p.b = [10, 11, 12, 13, 14] | units.m
        p.c = [20, 21, 22, 23, 24] | units.m
        
        io.write_set_to_file(
            p, 
            "test.csv",
            "txt", 
            attribute_names = ('a', 'b', 'c'),
            attribute_types = (None, units.m, units.m),
            maximum_number_of_lines_buffered = 1,
        )
        with open("test.csv", "r") as f:
            contents = f.read()
            
        expected_contents = '#a b c\n#- m m\n1.0 10.0 20.0\n2.0 11.0 21.0\n3.0 12.0 22.0\n4.0 13.0 23.0\n5.0 14.0 24.0\n'
        self.assertEqual(expected_contents, contents)
        p2 = io.read_set_from_file(
            "test.csv",
            "txt", 
            attribute_names = ('a', 'b', 'c'),
            attribute_types = (None, units.m, units.m),
            maximum_number_of_lines_buffered = 1,
        )
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)
        self.assertAlmostRelativeEquals(p2.c, p.c)
        
    def test10(self):
        p = datamodel.Particles(keys=[30,31,32,33,34])
        p.a = [1.0, 2.0, 3.0, 4.0, 5.0]
        p.b = [10, 11, 12, 13, 14] | units.m
        p.c = [20, 21, 22, 23, 24] | units.m
        io.write_set_to_file(
            p, 
            "test.csv",
            "txt", 
            attribute_names = ('a', 'b', 'c'),
            attribute_types = (None, units.m, units.m),
            maximum_number_of_lines_buffered = 1,
            key_in_column = 0
        )
        with open("test.csv", "r") as f:
            contents = f.read()
        expected_contents = '#a b c\n#- m m\n30 1.0 10.0 20.0\n31 2.0 11.0 21.0\n32 3.0 12.0 22.0\n33 4.0 13.0 23.0\n34 5.0 14.0 24.0\n'
        self.assertEqual(expected_contents, contents)
        p2 = io.read_set_from_file(
            "test.csv",
            "txt", 
            attribute_names = ('a', 'b', 'c'),
            attribute_types = (None, units.m, units.m),
            maximum_number_of_lines_buffered = 1,
            key_in_column = 0
        )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)
        self.assertAlmostRelativeEquals(p2.c, p.c)
        
    def test11(self):
        p = datamodel.Particles(200)
        p.a = 2 | units.m
        
        io.write_set_to_file(
            p, 
            "test.csv",
            "txt", 
            attribute_names = ('a'),
            attribute_types = (units.m,),
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        p2 = io.read_set_from_file(
            "test.csv",
            "txt", 
            attribute_names = ('a'),
            attribute_types = (units.m,),
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        
    def test12(self):
        print("Test Text IO with specific data types (string, int, float)")
        daltons = datamodel.Particles(keys=[30,31,32,33])
        daltons.name = ["Joe", "William", "Jack", "Averell"]
        daltons.length = [1.1, 1.4, 1.7, 2.0] | core.unit_with_specific_dtype(units.m, "float32")
        daltons.age = [21, 20, 19, 18] | core.unit_with_specific_dtype(units.yr, "int32")
        path=os.path.abspath(os.path.join(self.get_path_to_results(), "daltons.txt"))
        io.write_set_to_file(
            daltons, 
            path,
            "txt", 
            attribute_names = ('name', 'length', 'age'),
            attribute_types = (None, units.m, units.yr),
            maximum_number_of_lines_buffered = 2,
            key_in_column = 0
        )
        with open(path, "r") as f:
            contents = f.read()
        expected_contents = '#name length age\n#- m yr\n30 Joe 1.1 21\n31 William 1.4 20\n32 Jack 1.7 19\n33 Averell 2.0 18\n'
        self.assertEqual(expected_contents, contents)
        
        read = io.read_set_from_file(
            path,
            "txt", 
            attribute_names = ('name', 'length', 'age'),
            attribute_types = (None, units.m, units.yr),
            attribute_dtypes = ("str", "float32", "int32"),
            maximum_number_of_lines_buffered = 2,
            key_in_column = 0
        )
        self.assertEqual(read.key, daltons.key)
        self.assertEqual(read.name, daltons.name)
        self.assertEqual(read.length, daltons.length)
        self.assertEqual(read.age, daltons.age)
        self.assertTrue(read.name.dtype.kind in ["S", "U"])
        self.assertEqual(str(read.length.value_in(units.m).dtype), "float32")
        self.assertEqual(str(read.age.value_in(units.yr).dtype), "int32")
    

    def test13(self):
        p = datamodel.Particles(100)
        p.a = numpy.arange(0,1,0.01) | units.m
        
        path=os.path.abspath(os.path.join(self.get_path_to_results(), "test.csv"))

        io.write_set_to_file(
            p, 
            path,
            "amuse-txt", 
            attribute_names = ('a'),
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        p2 = io.read_set_from_file(
            path,
            "txt", 
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        


    def test14(self):
        p = datamodel.Particles(100)
        p.a = numpy.arange(0,1,0.01) | units.m
        p.b = numpy.arange(0,1,0.01)     
        
        io.write_set_to_file(
            p, 
            "test.csv",
            "amuse-txt", 
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        p2 = io.read_set_from_file(
            "test.csv",
            "txt", 
            maximum_number_of_lines_buffered = 10,
            key_in_column = 0
        )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)
        



class CsvFileTextTests(amusetest.TestCase):
    
    
    def test1(self):
        print("Test 1: Read comma separated values (CSV) - specified attributes")
        contents = "#header\n1,2,3\n4,5,6\n7,8,9\n"
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        instance.attribute_names = ['a', 'b', 'c']
        instance.attribute_types = [units.none, units.m, units.m/units.s]
        particles = instance.load()
        self.assertEqual(len(particles), 3)
        self.assertEqual(particles.a, [1, 4, 7] | units.none)
        self.assertEqual(particles.b, [2, 5, 8] | units.m)
        self.assertEqual(particles.c, [3, 6, 9] | units.m / units.s)
    
    def test2(self):
        print("Test 2: Read comma separated values (CSV) - attributes defined in header")
        contents = ("#a, b, c\n#no_system.get('none'),system.get('S.I.').base('length'),"
            "(system.get('S.I.').base('length') / system.get('S.I.').base('time'))\n"
            "#none, m, m/s\n1,2,3\n4,5,6\n7,8,9\n")
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        particles = instance.load()
        self.assertEqual(len(particles), 3)
        self.assertEqual(particles.a, [1, 4, 7] | units.none)
        self.assertEqual(particles.b, [2, 5, 8] | units.m)
        self.assertEqual(particles.c, [3, 6, 9] | units.m / units.s)
    
    def test3(self):
        print("Test 3: Read comma separated values (CSV) - generic units")
        contents = ("#a,b,c\n"
            "#system.get('generic').base('mass'),system.get('generic').base('length'),"
            "(((system.get('generic').base('length')**2) * (system.get('generic').base('time')**-2)) * "
            "system.get('generic').base('mass'))\n"
            "#mass,length,length**2 * time**-2 * mass\n1.0,2.0,3.0\n4.0,5.0,6.0\n")
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        particles = instance.load()
        self.assertEqual(len(particles), 2)
        self.assertEqual(particles.a, [1, 4] | generic_unit_system.mass)
        self.assertEqual(particles.b, [2, 5] | generic_unit_system.length)
        self.assertEqual(particles.c, [3, 6] | generic_unit_system.energy)
        
    def test4(self):
        print("Test 4: Write comma separated values (CSV) - specified attributes")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.attribute_names = ['a', 'b']
        instance.attribute_types = [units.none, 100*units.cm]
        instance.store()
        
        contents = data_stream.getvalue()
        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,100 * cm\n1.0,2.0\n4.0,5.0\n", contents)
    
    def test5(self):
        print("Test 5: Write comma separated values (CSV) - attributes defined automatically")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.store()
        
        contents = data_stream.getvalue()
        self.assertEqual("#a,b,c\n"
            "#no_system.get('none'),system.get('S.I.').base('length'),"
            "((system.get('S.I.').base('length')**-3) * system.get('S.I.').base('mass'))\n"
            "#none,m,m**-3 * kg\n1.0,2.0,3.0\n4.0,5.0,6.0\n", contents)
    
    def test6(self):
        print("Test 6: Write comma separated values (CSV) - generic units")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | generic_unit_system.mass
        particles.b = [2, 5] | generic_unit_system.length
        particles.c = [3, 6] | generic_unit_system.energy
        particles.d = [4, 7] | generic_unit_system.temperature
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.store()
        
        contents = data_stream.getvalue()
        self.assertEqual("#a,b,c,d\n"
            "#system.get('generic').base('mass'),system.get('generic').base('length'),"
            "(((system.get('generic').base('length')**2) * (system.get('generic').base('time')**-2)) * "
            "system.get('generic').base('mass')),system.get('generic').base('thermodynamic temperature')\n"
            "#mass,length,length**2 * time**-2 * mass,thermodynamic temperature\n1.0,2.0,3.0,4.0\n4.0,5.0,6.0,7.0\n", contents)
    
    def test7(self):
        print("Test 7: Write CSV - quantities instead of set, names and types unspecified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.store()
        self.assertEqual("#col(0),col(1)\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,m\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())
    
    def test8(self):
        print("Test 8: Write CSV - quantities instead of set, types unspecified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.attribute_names = ['a', 'b']
        instance.store()
        
        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,m\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())
    
    def test9(self):
        print("Test 9: Write CSV - quantities instead of set, names and types specified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m
        
        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.attribute_names = ['a', 'b']
        instance.attribute_types = [units.none, 100*units.cm]
        instance.store()
        
        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,100 * cm\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())
    
    def test10(self):
        print("Test 10: User interface (write_set_to_file and read_set_from_file)")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        io.write_set_to_file(particles, "test_textio.csv","csv")
        
        read_particles = io.read_set_from_file("test_textio.csv", format = "csv")
        self.assertEqual(len(read_particles), 2)
        self.assertEqual(read_particles.a, [1, 4] | units.none)
        self.assertEqual(read_particles.b, [2, 5] | units.m)
        self.assertEqual(read_particles.c, [3, 6] | units.kg / units.m**3)
        os.remove("test_textio.csv")
    
    
    def test11(self):
        particles = datamodel.Particles(2)
        particles.a = [1, 4]
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        io.write_set_to_file(
            particles, 
            "test_textio.csv",
            "csv",
            attribute_type=(None, units.kg / units.m**3, units.m),
            attribute_name=('a', 'c', 'b')
        )
        
        read_particles = io.read_set_from_file(
            "test_textio.csv", 
            format = "csv"
        )
        self.assertEqual(len(read_particles), 2)
        self.assertEqual(read_particles.a, [1, 4])
        self.assertEqual(read_particles.b, [2, 5] | units.m)
        self.assertEqual(read_particles.c, [3, 6] | units.kg / units.m**3)
        os.remove("test_textio.csv")

class Athena3DTextTests(amusetest.TestCase):
    
    def test1(self):
        contents = """\
        # Nx1 = 128
        # x1-size = 1
        # Time = 0.103125
        #
        # [1]=i-zone [2]=x1 [3]=d [4]=M1 [5]=M2 [6]=M3 [7]=P [8]=E [9]=B1c [10]=B2c [11]=B3c [12]=B1i [13]=B1i [14]=B1i
        #
          4  3.90625e-03  1.00000e+00 -8.66511e-07  4.08477e-07  1.44419e-07  6.00000e-01  2.52500e+00  1.00000e+00  1.41421e+00  5.00000e-01  1.00000e+00  1.41421e+00  5.00000e-01
        """
        data_file = StringIO(textwrap.dedent(contents))
        instance = text.Athena3DText("test.tab", data_file)
        particles = instance.load()
        
        self.assertEqual(len(particles), 1)
        self.assertEqual(particles[0].x1, 3.90625e-03 |units.none)
        self.assertEqual(particles[0].i_zone, 4 |units.none)
        self.assertEqual(particles[0].M1, -8.66511e-07 |units.none)
        
        
        
        
        
        
